%%% createMultipleMappingGoals
%%%		Action: goals based on multiple mappings are created

createMultipleMappingGoals:-
		findall((WS,M),(nounMapping2AdimenSUMO(WS,M),length(M,N),N>1),LNM),
		createGoalsListOfMultipleMappings(LNM,0,I1),
		findall((WS,M),(verbMapping2AdimenSUMO(WS,M),length(M,N),N>1),LVM),
		createGoalsListOfMultipleMappings(LVM,I1,I2),
		findall((WS,M),(adjMapping2AdimenSUMO(WS,M),length(M,N),N>1),LAM),
		createGoalsListOfMultipleMappings(LAM,I2,I3),
		findall((WS,M),(advMapping2AdimenSUMO(WS,M),length(M,N),N>1),LRM),
		createGoalsListOfMultipleMappings(LRM,I3,NI),
		setNumberOfGoals('multipleMapping',NI),
		setNumberOfGoals('negatedMultipleMapping',NI).

%%% createGoalsListOfMultipleMappings(in LM, in I, out NI)
%%%		LM: list of multiple mappings
%%%		I: index of goals
%%%		NI: new index of goals
%%%		Action: the goals based on LM are created

createGoalsListOfMultipleMappings([],I,I).
createGoalsListOfMultipleMappings([(WS,M)|LM],I,NI):-
		createMultipleMappingGoal(WS,M,I,I1),
		createGoalsListOfMultipleMappings(LM,I1,NI).

%%% createMultipleMappingGoal(in WS, in M, in I, out NI)
%%%		WS: synset
%%%		M: multiple mapping
%%%		I: index of goals
%%%		NI: new index of goals
%%%		Action: the goal based on M is created

createMultipleMappingGoal(WS,M,I,NI):-
		getMultipleMappingFormulas(M,LF),
		saveMultipleMappingformulas(WS,M,LF,I,NI).

%%% getMultipleMappingFormulas(in M, out LF)
%%%		M: mapping
%%%		LF: the list of formulas that corresponds to M

getMultipleMappingFormulas(M,LF):-
		classifyMultipleMappings(M,LO,SC,SR,EA,SA),
		sort(LO,SLO),
		sort(SC,SSC),
		sort(SR,SSR),
		sort(EA,SEA),
		sort(SA,SSA),
		findall(F,createMultipleMappingFormulas(SLO,SSC,SSR,SEA,SSA,F),LF).

%%% classifyMultipleMappings(in M, out LO, out SC, out SR, out EA, out SA)
%%%		M: multiple mapping
%%%		LO: list of objects that corresponds to M
%%%		SC: list of classes that corresponds to M
%%%		SR: list of relations that corresponds to M
%%%		EA: list of equivalent attributes that corresponds to M
%%%		SA: list of attributes that corresponds to M

classifyMultipleMappings([],[],[],[],[],[]).
classifyMultipleMappings([(C,equivalence)|M],[C|LO],SC,SR,EA,SA):-
		isRelation(C),
		!,
		classifyMultipleMappings(M,LO,SC,SR,EA,SA).
classifyMultipleMappings([(C,equivalence)|M],LO,SC,SR,[C|EA],SA):-
		isAttribute(C),
		!,
		classifyMultipleMappings(M,LO,SC,SR,EA,SA).
classifyMultipleMappings([(C,equivalence)|M],LO,[C|SC],SR,EA,SA):-
		isClass(C),
		!,
		print_to_screen('\nWarning --> Multiple mappings with equivalence to class'),
		classifyMultipleMappings(M,LO,SC,SR,EA,SA).
classifyMultipleMappings([(C,_)|M],LO,SC,[C|SR],EA,SA):-
		isRelation(C),
		!,
		classifyMultipleMappings(M,LO,SC,SR,EA,SA).
classifyMultipleMappings([(C,_)|M],LO,SC,SR,EA,[C|SA]):-
		isAttribute(C),
		!,
		classifyMultipleMappings(M,LO,SC,SR,EA,SA).
classifyMultipleMappings([(C,_)|M],[C|LO],SC,SR,EA,SA):-
		isObject(C),
		!,
		classifyMultipleMappings(M,LO,SC,SR,EA,SA).
classifyMultipleMappings([(C,_)|M],LO,[C|SC],SR,EA,SA):-
		isClass(C),
		!,
		classifyMultipleMappings(M,LO,SC,SR,EA,SA).

%%% createMultipleMappings(in LO, in SC, in SR, in EA, in SA, out F)
%%%		LO: list of objects
%%%		LC: list of classes
%%%		LR: list of relations
%%%		LEA: list of equivalent attributes
%%%		LA: list of attributes
%%%		F: formula that is obtained from LO, LC, LR, LEA and LA

/*
createMultipleMappingFormulas(LO,LC,LR,LEA,LA,F):-
		LEA \= [],
		X =..['#var','X'],
		Y =..['#var','Y'],
		append(LO,LEA,NLO),
		findall(SE,(member(O,NLO),SE=equal(X,O)),LSE),
		findall(SI,(member(I,LC),SI='$instance'(X,I)),LSI),
		(
		 (
		  findall(O,(member(O,LO),isInstanceOf(O,'Object')),PLO),
		  findall(I,(member(I,LC),isSubclassOf(I,'Object')),PLI),
		  append(LO,LC,LOC),
		  append(PLO,PLI,PLOI),
		  length(LOC,NOI),
		  length(PLOI,NOI),
		  !,
		  P = attribute
		 );
		 (
		  P = property
		 )
		),
		findall(SR,(member(R,LR),SR=subrelation(X,R)),LSR),
		findall(SA,(member(A,LA),PA=..[P,X,Y],SA=exists([Y],and([subAttribute(Y,A),PA]))),LSA),
		append([LSE,LSI,LSR,LSA],LF),
		(
		 (
		  length(LF,N),
		  N > 1,
		  !,
		  SF = and(LF)
		 );
		 (
		  SF = LF
		 )
		),
		F = exists([X],SF).
*/
createMultipleMappingFormulas(LO,LC,LR,LEA,LA,F):-
		LEA \= [],
		X =..['#var','X'],
		Y =..['#var','Y'],
		Z =..['#var','Z'],
		findall(SE,(member(O,LO),SE=equal(X,O)),LSE),
		findall(SI,(member(I,LC),SI='$instance'(X,I)),LSI),
		(
		 (
		  findall(O,(member(O,LO),isInstanceOf(O,'Object')),PLO),
		  findall(I,(member(I,LC),isSubclassOf(I,'Object')),PLI),
		  append(LO,LC,LOC),
		  append(PLO,PLI,PLOI),
		  length(LOC,NOI),
		  length(PLOI,NOI),
		  !,
		  P = attribute
		 );
		 (
		  P = property
		 )
		),
		findall(SR,(member(R,LR),SR=subrelation(X,R)),LSR),
		findall(SEA,(member(EA,LEA),SEA=..[P,X,EA]),LSEA),
		findall(SA,(member(A,LA),isObjectAttribute(A),PA=..[P,X,Y],SA=exists([Y],and([subAttribute(Y,A),PA]))),LSA1),
		findall(SA,(member(A,LA),isClassAttribute(A),PA=..[P,X,Y],SA=exists([Y,Z],and(['$instance'(Z,A),subAttribute(Y,Z),PA]))),LSA2),
		append([LSE,LSI,LSR,LSEA,LSA1,LSA2],LF),
		(
		 (
		  length(LF,N),
		  N > 1,
		  !,
		  SF = and(LF)
		 );
		 (
		  SF = LF
		 )
		),
		F = exists([X],SF).
createMultipleMappingFormulas(LO,LC,LR,LEA,LA,F):-
		LEA = [],
		X =..['#var','X'],
		Y =..['#var','Y'],
		Z =..['#var','Z'],
		findall(SE,(member(O,LO),SE=equal(X,O)),LSE),
		findall(SI,(member(I,LC),SI='$instance'(X,I)),LSI),
		(
		 (
		  findall(O,(member(O,LO),isInstanceOf(O,'Object')),PLO),
		  findall(I,(member(I,LC),isSubclassOf(I,'Object')),PLI),
		  append(LO,LC,LOC),
		  append(PLO,PLI,PLOI),
		  length(LOC,NOI),
		  length(PLOI,NOI),
		  !,
		  P = attribute
		 );
		 (
		  P = property
		 )
		),
		findall(SR,(member(R,LR),SR=subrelation(X,R)),LSR),
		findall(SA,(member(A,LA),isObjectAttribute(A),PA=..[P,X,Y],SA=exists([Y],and([subAttribute(Y,A),PA]))),LSA1),
		findall(SA,(member(A,LA),isClassAttribute(A),PA=..[P,X,Y],SA=exists([Y,Z],and(['$instance'(Z,A),subAttribute(Y,Z),PA]))),LSA2),
		append([LSE,LSI,LSR,LSA1,LSA2],LF),
		(
		 (
		  length(LF,N),
		  N > 1,
		  !,
		  SF = and(LF)
		 );
		 (
		  SF = LF
		 )
		),
		F = exists([X],SF).

%%% saveMultipleMappingFormula(in WS, in M, in LF, in I, out NI)
%%%		WS: synset
%%%		M: multiple mapping
%%%		LF: the list of formulas that corresponds to M
%%%		I: index of goals
%%%		NI: new index of goals
%%%		Action: the goals based on M are stored

saveMultipleMappingformulas(_,_,[],I,I).
saveMultipleMappingformulas(WS,M,[F|LF],I,NI):-
		getGoal(K1,not(F)),
		!,
		setGoalInformation(K1,mapping(WS,M),_,_,'multipleMapping'),
		getGoal(K2,F),
		setGoalInformation(K2,not(mapping(WS,M)),_,_,'negatedMultipleMapping'),
		saveMultipleMappingformulas(WS,M,LF,I,NI).
saveMultipleMappingformulas(WS,M,[F|LF],I,NI):-
		I1 is I+1,
		getGoalKey('multipleMapping',I1,K1),
		saveGoal('multipleMapping',I1,I1,K1,not(F)),
		setGoalInformation(K1,mapping(WS,M),_,_,'multipleMapping'),
		getGoalKey('negatedMultipleMapping',I1,K2),
		saveGoal('negatedMultipleMapping',I1,I1,K2,F),
		setGoalInformation(K2,not(mapping(WS,M)),_,_,'negatedMultipleMapping'),
		saveMultipleMappingformulas(WS,M,LF,I1,NI).
