% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% %																				% %
% %		Part IV:		Mapping to AdimenSUMO									% %
% %																				% %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %

%%% mapping2AdimenSUMO
%%%		Action: this predicate obtain the mapping to AdimenSUMO from the mapping
%%%					of SUMO

mapping2AdimenSUMO:-
		print_to_screen('\n'),
		print_to_screen('\tNouns...'),
		getNounMapping2AdimenSUMO,
		print_to_screen('\n'),
		print_to_screen('\tVerbs...'),
		getVerbMapping2AdimenSUMO,
		print_to_screen('\n'),
		print_to_screen('\tAdjectives...'),
		getAdjectiveMapping2AdimenSUMO,
		print_to_screen('\n'),
		print_to_screen('\tAdverbs...'),
		getAdverbMapping2AdimenSUMO.

%%% getNounMapping2AdimenSUMO
%%%		Action: this predicate obtain the mapping of nouns to AdimenSUMO from
%%%					the mapping of SUMO

getNounMapping2AdimenSUMO:-
		findall(('noun',WS,M),nounMapping2SUMO(WS,M),LNM),
		fromSUMOtoAdimenSUMOMapping(LNM),
		writeMapping2AdimenSUMO('noun').

%%% getVerbMapping2AdimenSUMO
%%%		Action: this predicate obtain the mapping of verbs to AdimenSUMO from
%%%					the mapping of SUMO

getVerbMapping2AdimenSUMO:-
		findall(('verb',WS,M),verbMapping2SUMO(WS,M),LVM),
		fromSUMOtoAdimenSUMOMapping(LVM),
		writeMapping2AdimenSUMO('verb').

%%% getAdjectiveMapping2AdimenSUMO
%%%		Action: this predicate obtain the mapping of adjectives to AdimenSUMO from
%%%					the mapping of SUMO

getAdjectiveMapping2AdimenSUMO:-
		findall(('adj',WS,M),adjMapping2SUMO(WS,M),LAM),
		fromSUMOtoAdimenSUMOMapping(LAM),
		writeMapping2AdimenSUMO('adj').

%%% getAdverbMapping2AdimenSUMO
%%%		Action: this predicate obtain the mapping of adverbs to AdimenSUMO from
%%%					the mapping of SUMO

getAdverbMapping2AdimenSUMO:-
		findall(('adv',WS,M),advMapping2SUMO(WS,M),LRM),
		fromSUMOtoAdimenSUMOMapping(LRM),
		writeMapping2AdimenSUMO('adv').

%%% fromSUMOtoAdimenSUMOMapping(in LM)
%%%		LM:  list of 3-tuples (T,WS,M) where
%%%				T: type of WS (noun, verb, adjective or adverb)
%%%				WS: WordNet synset
%%%				M: mapping to WS
%%%		Action: the list of mappings LM is translated into AdimenSUMO

fromSUMOtoAdimenSUMOMapping([]).
fromSUMOtoAdimenSUMOMapping([(T,WS,M)|LM]):-
		findall(NM,(member((C,R),M),fromSUMOtoAdimenSUMO(WS,C,R,NM)),LNM),
		append(LNM,LNM1),
		removeRedundantMappings(LNM1,LNM2),
		saveAdimenSUMOMapping(T,WS,LNM2),
		fromSUMOtoAdimenSUMOMapping(LM).

%%% fromSUMOtoAdimenSUMO(in WS, in C, in R, out M)
%%%		WS: WordNet synset
%%%		C: SUMO concept
%%%		R: mapping relation from WS to C
%%%		M: mapping from WS to AdimenSUMO obtained from (C,R)

fromSUMOtoAdimenSUMO(WS,C,R,[(C,NR)]):-
		isAdimenSUMO(C),
		!,
		(
		 (
		  R = undefined,
		  !,
		  NR = subsumption,
		  assertWithoutRepetition('&missingMappingRelation'(WS),_)
		 );
		 (
		  NR = R
		 )
		).
fromSUMOtoAdimenSUMO(WS,C,not(R),[('Entity',subsumption)]):-
		!,
		assertWithoutRepetition('&missingMapping'(WS,('Negative mapping',C,not(R))),_).
fromSUMOtoAdimenSUMO(WS,C,R,M):-
		isSUMO(C),
		!,
		(
		 (
		  R = equivalence,
		  !,
		  NR = subsumption
		 );
		 (
		  R = undefined,
		  !,
		  NR = subsumption,
		  assertWithoutRepetition('&missingMappingRelation'(WS),_)
		 );
		 (
		  NR = R
		 )
		),
		getAllPossibleMappings(C,NR,M),
		(
		 (
		  M = [],
		  !,
		  assertWithoutRepetition('&missingMapping'(WS,('Broken link from SUMO to AdimenSUMO',C,R)),_)
		 );
		 (
		  true
		 )
		).
fromSUMOtoAdimenSUMO(WS,C,R,[('Entity',subsumption)]):-
		!,
		assertWithoutRepetition('&missingMapping'(WS,('Non-existing SUMO concept',C,R)),_).

%%% saveAdimenSUMOMapping(in T, in WS, in M)
%%%		T: type of synset
%%%		WS: WordNet synset
%%%		M: mapping to WS
%%%		Action: the mapping to AdimenSUMO is stores using dynamic predicates

saveAdimenSUMOMapping(T,WS,M):-
		atom_concat(T,'Mapping2AdimenSUMO',P),
		MA =..[P,WS,M],
		assertz((MA)).

%%% getAllPossibleMappings(in C, in R, out LM)
%%%		C: SUMO concept
%%%		R: mapping relation
%%%		LM: list of all possible mappings to AdimenSUMO obtained from C and R

getAllPossibleMappings(C,R,NLM):-
		getAllDirectSuperConceptsOf(C,LSC),
		remove_duplicates(LSC,NLSC),
		getPossibleMappings(NLSC,R,LM),
		remove_duplicates(LM,NLM).

%%% getPossibleMappings(in LC, in R, out LM)
%%%		LC: list of SUMO concept
%%%		R: mapping relation
%%%		LM: list of all possible mappings obtained from LC and R

getPossibleMappings([],_,[]).
getPossibleMappings([C|LC],R,[(C,R)|LM]):-
		isAdimenSUMO(C),
		!,
		getPossibleMappings(LC,R,LM).
getPossibleMappings([C|LC],R,LM):-
		isSUMO(C),
		!,
		getAllDirectSuperConceptsOf(C,LSC),
		append(LSC,LC,NLC),
		getPossibleMappings(NLC,R,LM).
getPossibleMappings([_|LC],R,LM):-
		getPossibleMappings(LC,R,LM).

%%% removeRedundantMappings(in M, out NM)
%%%		M: mapping
%%%		NM: the mapping that is obtained from M by removing redundant mappings

removeRedundantMappings(M,NM):-
		removeRedundantMappings([],M,NM).

%%% removeRedundantMappings(in M1, out M2, out NM)
%%%		M1: submapping already treated
%%%		M2: submapping to be treated
%%%		NM: the mapping that is obtained from M1 and M2 by removing redundant mappings

removeRedundantMappings(M,[],M).
removeRedundantMappings(M1,[M2|LM2],NM):-
		removeRedundantSingleMapping(M2,M1,_,NM1),
		removeRedundantSingleMapping(M2,LM2,NM2,NLM2),
		append(NM1,NM2,NNM1),
		removeRedundantMappings(NNM1,NLM2,NM).

%%% removeRedundantSingleMapping(in SM, in M, out NSM, out NM)
%%%		SM: single mapping
%%%		M: mapping
%%%		NSM: empty list (if SM is redundant) or SM
%%%		NM: the mapping that is obtained from M by removing the mappings redundant with SM

removeRedundantSingleMapping(M,[],[M],[]).
removeRedundantSingleMapping((C,R),[(C,R)|M],NSM,NM):-
		!,
		removeRedundantSingleMapping((C,R),M,NSM,NM).
removeRedundantSingleMapping((C,R1),[(C,R2)|M],NSM,NM):-
		!,
		(
		 (
		  R1 = equivalence,
		  !,
		  NR = equivalence
		 );
		 (
		  R2 = equivalence,
		  !,
		  NR = equivalence
		 );
		 (
		  R1 = subsumption,
		  !,
		  NR = subsumption
		 );
		 (
		  R2 = subsumption,
		  !,
		  NR = subsumption
		 );
		 (
		  NR = instance
		 )
		),
		removeRedundantSingleMapping((C,NR),M,NSM,NM).
removeRedundantSingleMapping((C1,R1),[(C2,_)|M],NSM,NM):-
		(
		 (
		  isInstanceOf(C1,C2)
		 );
		 (
		  isSubclassOf(C1,C2)
		 );
		 (
		  isSubrelationOf(C1,C2)
		 );
		 (
		  isSubAttributeOf(C1,C2)
		 )
		),
		!,
		removeRedundantSingleMapping((C1,R1),M,NSM,NM).
removeRedundantSingleMapping((C1,_),[(C2,R2)|M],[],[(C2,R2)|M]):-
		(
		 (
		  isInstanceOf(C2,C1)
		 );
		 (
		  isSubclassOf(C2,C1)
		 );
		 (
		  isSubrelationOf(C2,C1)
		 );
		 (
		  isSubAttributeOf(C2,C1)
		 )
		),
		!.
removeRedundantSingleMapping((C1,R1),[(C2,R2)|M],NSM,[(C2,R2)|NM]):-
		removeRedundantSingleMapping((C1,R1),M,NSM,NM).

%%% writeMapping2AdimenSUMO(in T)
%%%		T: type of synset
%%%		Action: the mapping to AdimenSUMO is stored in a file

writeMapping2AdimenSUMO(T):-
		atom_concat(T,'Mapping2AdimenSUMO',P),
		atom_concat(P,'.pl',F),
		atom_concat('Mappings/',F,NF),
		open(NF,write,S),
		A =..[P,_,_],
		findall(_,(A,write_term(S,A,[quoted(true)]),print_to_file(S,'.\n')),_),
		close(S).
