%%% writeGoals
%%%		Action: write goals in the output files

writeGoals:-
		findall((F,N),numberOfGoals(F,N),LFN),
		catch(make_directory('Goals'),_,true),
		writeGoalsListOfFiles(LFN).

%%% writeGoalsListOfFiles(in LFN)
%%%		LFN: a list of pairs (F,N) where
%%%			F is the file name
%%%			N is the number of goals
%%%		Action: write goals in the output files

writeGoalsListOfFiles([]).
writeGoalsListOfFiles([(F,N)|LFN]):-
		atom_concat('Goals/',F,F1),
		atom_concat(F1,'.kif',NF),
		open(NF,write,S),
		writeGoalsFile(S,F,1,N,RG),
		close(S),
		print_to_screen('\n\t'),
		print_to_screen('File '),
		print_to_screen(NF),
		print_to_screen(' completed (total goals --> '),
		print_to_screen(RG),
		print_to_screen(').'),
		writeGoalsListOfFiles(LFN).

%%% writeGoalsFile(in S, in F, in CG, in NG, out NRG)
%%%		S: output stream
%%%		F: a file name
%%%		IG: index of goals
%%%		NG: total number of goals in the file
%%%		NRG: number of real goals that have been written to the file
%%%		Action: write automatically sorted goals in the output file

writeGoalsFile(_,_,IG,NG,0):-
		IG > NG,
		!.
writeGoalsFile(S,F,IG,NG,NRG):-
		getGoalKey(F,IG,K),
		getGoalInformation(K,LI),
		writeGoalInformation(S,LI,K),
		print_to_file(S,'\n'),
		getGoal(K,G),
		writeGoal(S,G,K,RG1),
		NIG is IG + 1,
		writeGoalsFile(S,F,NIG,NG,RG2),
		NRG is RG1 + RG2.

%%% writeGoalInformation(in S, in LI, in K)
%%%		S: output stream
%%%		LI: list of information of the goal, where each element is of the form (R,M1,MT1,M2,MT2,C)
%%%			R: role
%%%			M1,M2: mappings to AdimenSUMO
%%%			MT1,MT2: type of mappings to AdimenSUMO
%%%			C: code of the action
%%%		K: goal key
%%%		Action: the goal information is written in the output stream

writeGoalInformation(S,LI,K):-
		print_to_file(S,';; Goal key: '),
		print_to_file(S,K),
		print_to_file(S,'\n'),
		writeGoalListOfInformations(S,LI,K).

%%% writeGoalListOfInformations(in S, in LI, in K)
%%%		S: output stream
%%%		LI: list of information of the goal, where each element is of the form (R,M1,MT1,M2,MT2,C)
%%%			R: role
%%%			M1,M2: mappings to AdimenSUMO
%%%			MT1,MT2: type of mappings to AdimenSUMO
%%%			C: code of the action
%%%		K: goal key
%%%		Action: the goal information is written in the output stream

writeGoalListOfInformations(_,[],_):-
		!.
writeGoalListOfInformations(S,[(R,_,_,'multipleMapping')|LI],K):-
		print_to_file(S,';;\tRole: '),
		print_to_file(S,R),
%trace,
		R =..[_,SS|_],
		getSynsetSense(SS,SSS),
		print_to_file(S,'\n;;\t\t'),
		print_to_file(S,SS),
		print_to_file(S,' --> '),
		print_to_file(S,SSS),
		print_to_file(S,'\n'),
		!,
		writeGoalListOfInformations(S,LI,K).
writeGoalListOfInformations(S,[(not(R),_,_,'negatedMultipleMapping')|LI],K):-
		print_to_file(S,';;\tRole: '),
		print_to_file(S,R),
		R =..[_,SS|_],
		getSynsetSense(SS,SSS),
		print_to_file(S,'\n;;\t\t'),
		print_to_file(S,SS),
		print_to_file(S,' --> '),
		print_to_file(S,SSS),
		print_to_file(S,'\n'),
		!,
		writeGoalListOfInformations(S,LI,K).
writeGoalListOfInformations(S,[(R,M1,M2,C)|LI],K):-
		print_to_file(S,';;\tRole: '),
		print_to_file(S,R),
		R =..[_,S1,S2],
		getSynsetSense(S1,SS1),
		getSynsetSense(S2,SS2),
		print_to_file(S,'\n;;\t\t'),
		print_to_file(S,S1),
		print_to_file(S,' --> '),
		print_to_file(S,SS1),
		print_to_file(S,'\n;;\t\t'),
		print_to_file(S,S2),
		print_to_file(S,' --> '),
		print_to_file(S,SS2),
		print_to_file(S,'\n;;\t\tMapping of the 1st argument: '),
		print_to_file(S,M1),
		print_to_file(S,'\n;;\t\tMapping of the 2nd argument: '),
		print_to_file(S,M2),
		print_to_file(S,'\n;;\t\tAction: '),
		print_to_file(S,C),
		print_to_file(S,'\n'),
		!,
		writeGoalListOfInformations(S,LI,K).
writeGoalListOfInformations(_,_,K):-
		print_to_screen('Error writting the information of goal '),
		print_to_screen(K),
		print_to_screen('\n'),
		fail.

%%% writeGoal(in S, in OL, in AP, in G, in K, out NRA)
%%%		S: output stream
%%%		G: goal
%%%		K: goal key
%%%		NRA: number of real goals that have been written to the file
%%%		Action: the goal is written in the output stream

% % KIF format % %

writeGoal(S,G,_,1):-
		writeFormula(S,0,G,_),
		!,
		print_to_file(S,'\n').

% % Error % %

writeGoal(_,_,K,_):-
		print_to_screen('Error writting the goal '),
		print_to_screen(K),
		print_to_screen('\n'),
		fail.

%%% writeFormula(in S, in T, in F, out NUC)
%%%		S: output stream
%%%		T: number of tabs
%%%		F: a formula
%%%		NUC: number of unit clauses that have been written to the file (axioms consisting of a single literal)
%%%		Action: the formula is written in the output stream

writeFormula(S,T,exists(LV,F),_):-
		!,
		printTabs(S,T),
		print_to_file(S,'('),
		print_to_file(S,'exists '),
		print_to_file(S,'('),
		writeListOfVariables(S,LV),
		print_to_file(S,')'),
		print_to_file(S,'\n'),
		NT is T + 1,
		writeFormula(S,NT,F,_),
		printTabs(S,T),
		print_to_file(S,')'),
		print_to_file(S,'\n').
writeFormula(S,T,forall(LV,F),_):-
		!,
		printTabs(S,T),
		print_to_file(S,'('),
		print_to_file(S,'forall '),
		print_to_file(S,'('),
		writeListOfVariables(S,LV),
		print_to_file(S,')'),
		print_to_file(S,'\n'),
		NT is T + 1,
		writeFormula(S,NT,F,_),
		printTabs(S,T),
		print_to_file(S,')'),
		print_to_file(S,'\n').
writeFormula(S,T,exists*(LV,F),_):-
		!,
		printTabs(S,T),
		print_to_file(S,'('),
		print_to_file(S,'exists '),
		print_to_file(S,'('),
		writeListOfVariables(S,LV),
		print_to_file(S,')'),
		print_to_file(S,'\n'),
		NT is T + 1,
		writeFormula(S,NT,F,_),
		printTabs(S,T),
		print_to_file(S,')'),
		print_to_file(S,'\n').
writeFormula(S,T,forall*(LV,F),_):-
		!,
		printTabs(S,T),
		print_to_file(S,'('),
		print_to_file(S,'forall '),
		print_to_file(S,'('),
		writeListOfVariables(S,LV),
		print_to_file(S,')'),
		print_to_file(S,'\n'),
		NT is T + 1,
		writeFormula(S,NT,F,_),
		printTabs(S,T),
		print_to_file(S,')'),
		print_to_file(S,'\n').
writeFormula(S,T,and(LF),_):-
		!,
		printTabs(S,T),
		print_to_file(S,'('),
		print_to_file(S,'and '),
		print_to_file(S,'\n'),
		NT is T + 1,
		writeListOfFormulas(S,NT,LF),
		printTabs(S,T),
		print_to_file(S,')'),
		print_to_file(S,'\n').
writeFormula(S,T,or(LF),_):-
		!,
		printTabs(S,T),
		print_to_file(S,'('),
		print_to_file(S,'or '),
		print_to_file(S,'\n'),
		NT is T + 1,
		writeListOfFormulas(S,NT,LF),
		printTabs(S,T),
		print_to_file(S,')'),
		print_to_file(S,'\n').
writeFormula(S,T,implies(F1,F2),_):-
		!,
		printTabs(S,T),
		print_to_file(S,'('),
		print_to_file(S,'=> '),
		print_to_file(S,'\n'),
		NT is T + 1,
		writeFormula(S,NT,F1,_),
		writeFormula(S,NT,F2,_),
		printTabs(S,T),
		print_to_file(S,')'),
		print_to_file(S,'\n').
writeFormula(S,T,equiv(F1,F2),_):-
		!,
		printTabs(S,T),
		print_to_file(S,'('),
		print_to_file(S,'<=> '),
		print_to_file(S,'\n'),
		NT is T + 1,
		writeFormula(S,NT,F1,_),
		writeFormula(S,NT,F2,_),
		printTabs(S,T),
		print_to_file(S,')'),
		print_to_file(S,'\n').
writeFormula(S,T,not(F),_):-
		!,
		printTabs(S,T),
		print_to_file(S,'('),
		print_to_file(S,'not '),
		print_to_file(S,'\n'),
		NT is T + 1,
		writeFormula(S,NT,F,_),
		printTabs(S,T),
		print_to_file(S,')'),
		print_to_file(S,'\n').
writeFormula(S,T,A,_):-
		printTabs(S,T),
		writeAtom(S,T,A).

%%% writeListOfFormulas(in S, in LT, in LF)
%%%		S: output stream
%%%		T: number of tabs
%%%		LF: a list of formulas
%%%		Action: the formula is written in the output stream

% % KIF format % %

writeListOfFormulas(_,_,[]).
writeListOfFormulas(S,T,[F|LF]):-
		writeFormula(S,T,F,_),
		writeListOfFormulas(S,T,LF).

%%% writeAtom(in S, in T, in A)
%%%		S: output stream
%%%		T: number of tabs
%%%		A: an atom
%%%		Action: the atom is written in the output stream

writeAtom(S,_,false):-
		!,
		print_to_file(S,'false'),
		print_to_file(S,'\n').
writeAtom(S,_,true):-
		!,
		print_to_file(S,'true'),
		print_to_file(S,'\n').
writeAtom(S,T,A):-
		A =..[P|ST],
		print_to_file(S,'('),
		print_to_file(S,P),
		NT is T + 1,
		writeAtomSubterms(S,NT,ST),
		print_to_file(S,')'),
		print_to_file(S,'\n').

%%% writeAtomSubterms(in S, in T, in ST)
%%%		S: output stream
%%%		T: number of tabs
%%%		ST: a list of subterms
%%%		Action: the subterms of the atom are written in the output stream

writeAtomSubterms(_,_,[]):-
		!.
writeAtomSubterms(S,T,[H|ST]):-
		print_to_file(S,' '),
		writeTerm(S,T,H),
		NT is T + 1,
		writeAtomSubterms(S,NT,ST).

%%% writeTerm(in S, in T, in TE)
%%%		S: output stream
%%%		T: number of tabs
%%%		TE: a term
%%%		Action: the term is written in the output stream

writeTerm(S,_,'#var'(V)):-
		!,
		writeVariable(S,V).
writeTerm(S,_,'#rowVar'(V)):-
		!,
		writeRowVariable(S,V).
writeTerm(S,_,'#row'(L)):-
		!,
		writeMultiLengthList(S,L).
writeTerm(S,_,'#pred'(P)):-
		!,
		writeVariablePredicate(S,P).
writeTerm(S,T,A):-
		compound(A),
		!,
		A =..[F|ST],
		length(ST,N),
		print_to_file(S,'('),
		writeFunction(S,T,F,N),
		print_to_file(S,' '),
		writeSubterms(S,T,ST,' '),
		print_to_file(S,')').
writeTerm(S,T,A):-
		writeConstant(S,T,A).

%%% writeConstant(in S, in T, in CO)
%%%		S: output stream
%%%		T: number of tabs
%%%		CO: a constant
%%%		Action: the constant is written in the output stream

writeConstant(S,_,P):-
		!,
		print_to_file(S,P).

%%% writeFunction(in S, in T, in F, in A)
%%%		S: output stream
%%%		T: number of tabs
%%%		F: a function symbol
%%%		A: an arity
%%%		Action: the function is written in the output stream

writeFunction(S,_,F,_):-
		!,
		print_to_file(S,F).

%%% writeSubterms(in S, in T, in ST, in D)
%%%		S: output stream
%%%		T: number of tabs
%%%		ST: a list of subterms
%%%		D: the separator of subterms
%%%		Action: the function is written in the output stream

writeSubterms(S,T,[ST],_):-
		!,
		writeTerm(S,T,ST).
writeSubterms(S,T,[ST|LST],D):-
		writeTerm(S,T,ST),
		print_to_file(S,D),
		writeSubterms(S,T,LST,D).

%%% writeListOfVariables(in S,in LV)
%%%		S: output stream
%%%		LV: a list of variables
%%%		Action: the variable is written in the output stream

writeListOfVariables(_,[]).
writeListOfVariables(S,['#var'(V)]):-
		!,
		writeVariable(S,V).
writeListOfVariables(S,['#rowVar'(V)]):-
		!,
		writeRowVariable(S,V).
writeListOfVariables(S,['#var'(V)|LV]):-
		!,
		writeVariable(S,V),
		print_to_file(S,' '),
		writeListOfVariables(S,LV).
writeListOfVariables(S,['#rowVar'(V)|LV]):-
		writeRowVariable(S,V),
		print_to_file(S,' '),
		writeListOfVariables(S,LV).

%%% writeVariable(in S, in V)
%%%		S: output stream
%%%		V: a variable
%%%		Action: the variable is written in the output stream

writeVariable(S,V):-
		print_to_file(S,'?'),
		print_to_file(S,V).

%%% writeRowVariable(in S, in V)
%%%		S: output stream
%%%		V: a row variable
%%%		Action: the row variable is written in the output stream

writeRowVariable(S,V):-
		print_to_file(S,'@'),
		print_to_file(S,V).

%%% writeMultiLengthList(in S, in L)
%%%		S: output stream
%%%		L: a multi-length list
%%%		Action: the multi-length list is written in the output stream

writeMultiLengthList(S,[E|L]):-
		print_to_file(S,'@('),
		writeListOfSubterms(S,[E|L]),
		print_to_file(S,')').

writeMultiLengthList(S,T):-													% It is not a multi-length list. In fact, it is a row type
		print_to_file(S,'@'),
		print_to_file(S,T).

%%% writeListOfSubterms(in S, in L)
%%%		S: output stream
%%%		L: a list of terms
%%%		Action: the list of terms is written in the output stream

writeListOfSubterms(_,[]).
writeListOfSubterms(S,[E]):-
		!,
		writeTerm(S,_,E).
writeListOfSubterms(S,[E|L]):-
		!,
		writeTerm(S,_,E),
		print_to_file(S,','),
		writeListOfSubterms(S,L).

%%% writeVariablePredicate(in S, in V)
%%%		S: output stream
%%%		V: a variable predicate
%%%		Action: the variable predicate is written in the output stream

writeVariablePredicate(S,V):-
		print_to_file(S,'!'),
		print_to_file(S,V).

%%% printTabs(in S, in T)
%%%		S: output stream
%%%		T: a number of tabs
%%%		Action: T tabs are written in the output stream

printTabs(_,0):-
		!.
printTabs(S,T):-
		print_to_file(S,'\t'),
		NT is T - 1,
		printTabs(S,NT).

%%% getSynsetSense(in S, out SS)
%%%		S: synset
%%%		SS: sense of S

getSynsetSense(S,SS):-
		(
		 (
		  nounSense(S,SS)
		 );
		 (
		  verbSense(S,SS)
		 );
		 (
		  adjSense(S,SS)
		 );
		 (
		  advSense(S,SS)
		 )
		),
		!.

%%% print_to_file(in S, in T)
%%%		S: output strem
%%%		T: term
%%%		Action: the term is written in S

print_to_file(S,T):-
		write_term(S,T,[]).

%%% print_to_screen(in T)
%%%		T: term
%%%		Action: the term is prompted

print_to_screen(T):-
		current_output(S),
		print_to_file(S,T).
